{***********************************************************
 Component: TEditableCtrlHolder version 3.3

 TEditableCtrlHolder is a data validator of any 'TEditableCtrl'(1) controls of your form.
 For each of TEditableCtrl, you can set their DataType, MaxLength and Required property.
 Check the validity of what you entered into each TEditableCtrl against the properties defined above.
 You can either set a custom validation rule with the OnValidate event-handler that returns the Accept: boolean variable.
 You can also loop into all TEditableCtrl to check if the entered values have been modified
(1): a TEditableCtrl can be any of editable controls, ie TEdit, TComboBox, TListBox, TCheckBox, TMemo, ..

 Author: Didier Cabal
         'Bellefont'
         31800 Saint-Gaudens, France
         tl: +33 (0)5 61 89 72 92
         e-mail: didier.cabale+delphi@gmail.com
         web-page: http://didier.cabale.free.fr/delphi.htm

 history:
 version2.0: support RxLib TRxCheckListBox as TEditableCtrl
 version2.1: replaced 'null forbidden' by Required property
 version2.2: added function Validate: boolean;
              added property UnValidControl: TWinControl
 version2.3: added procedure RestoreInitialValue;
 version3.1: added OnValidate event-handler;
 version3.2: support TDateTimePicker as TEditableCtrl
 version3.3: support RxLib TDateEdit and TDBDateEdit as TEditableCtrl

************************************************************}

{$DEFINE RxLib}

unit EditableCtrlHolder;

interface

uses
   Classes,
   Sysutils,
   forms,
   controls,
   stdctrls,
   ComCtrls,
   Variants,
     {$IFDEF RxLib} RxCtrls, ToolEdit {$ENDIF}
     ;

type

  TDataType = (dtString, dtInteger, dtDouble, dtBoolean);
  TValidateResult = (vrOk, vrIsNull, vrWrongType, vrTooLong, vrWrongValue);
  TValidateProc = procedure(Sender: TWinControl; var Accept: boolean) of Object;

  TEditableCtrl = class(TCollectionItem)
  private
    FDataType: TDataType;
    FInitialValue: Variant;
    FRequired: boolean;
    FMaxLength: byte;
    FWinControl: TWinControl;
    FOnValidate: TValidateProc;
    function GetDisplayName: string; override;
    procedure SetWinControl(const Value: TWinControl);
//    procedure SetOnValidate(const Value: TEditableCtrlEvent);
  public
    function IsModifiedEntry: boolean;
    function ValidateEntry: TValidateResult;
    // when exported to sap_gui, this will have to return an integer -> messages.xml
    function GetValidateMessage(const VR: TValidateResult): string;
  published
    property WinControl: TWinControl read FWinControl write SetWinControl;
    property DataType: TDataType read FDataType write FDataType;
    property MaxLength: byte read FMaxLength write FMaxLength;
    property Required: boolean read FRequired write FRequired;
    property OnValidate: TValidateProc read FOnValidate write FOnValidate;
  end;

  TEditableCtrlHolder = class; // needs to be declared before because mutual recursion

  TEditableCtrlCollection = class(TCollection)
  private
    FEditableCtrlHolder: TEditableCtrlHolder;
    function GetItem(Index: Integer): TEditableCtrl;
    procedure SetItem(Index: Integer; Value: TEditableCtrl);
  protected
    function GetOwner: TPersistent; override;
  public
    constructor Create(aEditableCtrlHolder: TEditableCtrlHolder);
    function Add: TEditableCtrl;
    property Items[Index: Integer]: TEditableCtrl
      read GetItem write SetItem; default;
  end;

  TEditableCtrlHolder = class(TComponent)
  private
    FUnValidItem: TEditableCtrl;
    FItems: TEditableCtrlCollection;
    procedure SetItems(Value: TEditableCtrlCollection);
    function GetUnValidControl: TWinControl;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure LoadInitialValues;
    procedure RestoreInitialValues;
    function Validate: TValidateResult;
    property UnValidControl: TWinControl read GetUnValidControl;
  published
    property Items: TEditableCtrlCollection read FItems write SetItems;
  end;

resourcestring
   OkMsg         = 'Ok';
   RequiredMsg   = 'required';
   WrongTypeMsg  = 'wrong type';
   TooLongMsg    = 'too long';
   WrongValueMsg = 'wrong value';

procedure Register;

implementation

const
   BooToChar: array[Boolean] of Char = ('0', '1');
   CharToBoo: array['0'..'1'] of boolean = (false, true);

procedure Register;
begin
  RegisterComponents('DCComp', [TEditableCtrlHolder]);
end;


{ TEditableCtrl }


// Note: Inherited default behavior of GetDisplayName is to
// return the classname.

function TEditableCtrl.GetDisplayName: string;
begin
//  Result := Text;
  if Result = '' then Result := inherited GetDisplayName;
end;

function TEditableCtrl.GetValidateMessage(
  const VR: TValidateResult): string;
const
   ValidateMessages: Array[TValidateResult] of string = (OkMsg, RequiredMsg, WrongTypeMsg, TooLongMsg, WrongValueMsg);
begin
   result := ValidateMessages[VR];
end;

function TEditableCtrl.IsModifiedEntry: boolean;
var
   i: byte;
   temp: string;
begin
   result := true;
   if (FWinControl is TCustomEdit) and (FInitialValue = TCustomEdit(FWinControl).Text)
   or (FWinControl is TCustomComboBox) and (FInitialValue = TCustomComboBox(FWinControl).ItemIndex)
   or (FWinControl is TCheckBox) and (FInitialValue = TCheckBox(FWinControl).Checked)
   or (FWinControl is TDateTimePicker) and (FInitialValue = TDateTimePicker(FWinControl).Date) then
      result := false
//   else
   {$IFDEF RxLib}
   else if (FWinControl is TRxCheckListBox) then begin
      for i := 0 to TRxCheckListBox(FWinControl).Items.Count -1 do
         temp := temp + BooToChar[TRxCheckListBox(FWinControl).Checked[i]];
      if FInitialValue = temp then
         result := false;
   end;
   {$ENDIF}
end;

procedure TEditableCtrl.SetWinControl(const Value: TWinControl);
var
   i: byte;
begin
   FWinControl := Value;
{   if (Value is TCustomEdit) then begin
      FInitialValue := TCustomEdit(Value).Text;
   end else }if (Value is TCustomComboBox) then begin
      FDataType := dtInteger;
//      FInitialValue := TCustomComboBox(Value).ItemIndex;
   end else if (Value is TCheckBox) then begin
      FDataType := dtBoolean;
   end else if (Value is TDateTimePicker) then begin
      FDataType := dtDouble;
   {$IFDEF RxLib}
   end else if (Value is TRxCheckListBox) then begin
      FDataType := dtString;
{      FInitialValue := '';
      for i := 0 to TRxCheckListBox(Value).Items.Count -1 do
         FInitialValue := FInitialValue + s[TRxCheckListBox(Value).Checked[i]];}
   {$ENDIF}
   end;
end;

function TEditableCtrl.ValidateEntry: TValidateResult;
var
   value: string;
   v, code: integer;
   e: extended;
   checked,
   Accept: boolean;
   i: byte;
begin
   result := vrOk;
   if (FWinControl is TCustomEdit) then begin
      value := TCustomEdit(FWinControl).Text;

{      // allow null
      result := vrOk;
      if (not FRequired) and (Length(value) = 0) then
         exit;

      // check null forbidden
      result := vrRequired;
      if FRequired and (Length(value) = 0) then
         exit;}

      // check null forbidden
      if (Length(value) = 0) then
         if FRequired then
            result := vrIsNull
         else
//            result := vrOk
      else begin
         // check data type
         case FDataType of
         dtInteger: begin
                    Val(value, v, code);
                    if code <> 0 then
                        result := vrWrongType;
                    end;
         dtDouble:  if not TextToFloat(PChar(value), e, fvExtended) then
                    result := vrWrongType;
         dtBoolean: if not ((StrToBool(value) = true) or (StrToBool(value) = false)) then
                    result := vrWrongType;
         end;

      // check length
         if (FMaxLength > 0) and (System.Length(value) > FMaxLength) then
            result := vrTooLong; //ie for all other cases
//      result := vrOk;
      end;
//   end else if (FWinControl is TDateTimePicker) then begin
      // nothing to do except to set a custom OnValidate event-handler
   end;
   {$IFDEF RxLib}
   if (FWinControl is TCustomDateEdit) then begin
      result := vrOk;
      if (FRequired and (TCustomDateEdit(FWinControl).Date = 0)) then
         result := vrIsNull;
   end else if (FWinControl is TRxCheckListBox) then begin
      // check required
      result := vrOk;
      if FRequired then begin
         checked := false;
         for i := 0 to TRxCheckListBox(FWinControl).Items.Count -1 do
            checked := checked or TRxCheckListBox(FWinControl).Checked[i];
//      if TRxCheckListBox(FWinControl).SelCount = 0 then
         if not checked then
            result := vrIsNull;
      end;
   end;
   {$ENDIF}
   if result = vrOk then
//   Abort := false;
      if Assigned(FOnValidate) then begin
         Accept := true;
         FOnValidate(FWinControl, Accept);
         if not Accept then
            result := vrWrongValue;
      end;
end;


{ TEditableCtrlCollection }

constructor TEditableCtrlCollection.Create(aEditableCtrlHolder: TEditableCtrlHolder);
begin
   inherited Create(TEditableCtrl);
   FEditableCtrlHolder := aEditableCtrlHolder;
//   UnValidItem := nil;
end;

// Note: You must override GetOwner in Delphi 3.x to get
// correct streaming behavior.
function TEditableCtrlCollection.GetOwner: TPersistent;
begin
  Result := FEditableCtrlHolder;
end;

function TEditableCtrlCollection.Add: TEditableCtrl;
begin
  Result := TEditableCtrl(inherited Add);
end;

function TEditableCtrlCollection.GetItem(Index: Integer): TEditableCtrl;
begin
  Result := TEditableCtrl(inherited GetItem(Index));
end;

procedure TEditableCtrlCollection.SetItem(Index: Integer;
	Value: TEditableCtrl);
begin
  inherited SetItem(Index, Value);
end;


{ TEditableCtrlHolder }

constructor TEditableCtrlHolder.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FItems := TEditableCtrlCollection.Create(Self);
end;

destructor TEditableCtrlHolder.Destroy;
begin
  FItems.Free;
  inherited Destroy;
end;

function TEditableCtrlHolder.GetUnValidControl: TWinControl;
begin
   result := FUnValidItem.WinControl;
end;

procedure TEditableCtrlHolder.LoadInitialValues;
var
   i, j: byte;
   WinCtrl: TWinControl;
begin
   for i := 0 to Items.Count -1 do begin
   WinCtrl := Items[i].FWinControl;
   if (WinCtrl is TCustomEdit) then begin
      Items[i].FInitialValue := TCustomEdit(WinCtrl).Text;
   end else if (WinCtrl is TCustomComboBox) then begin
      Items[i].FInitialValue := TCustomComboBox(WinCtrl).ItemIndex;
   end else if (WinCtrl is TCheckBox) then begin
      Items[i].FInitialValue := TCheckBox(WinCtrl).Checked;
   end else if (WinCtrl is TDateTimePicker) then begin
      Items[i].FInitialValue := TDateTimePicker(WinCtrl).Date;
   {$IFDEF RxLib}
   end else if (WinCtrl is TRxCheckListBox) then begin
      Items[i].FInitialValue := '';
      for j := 0 to TRxCheckListBox(WinCtrl).Items.Count -1 do
         Items[i].FInitialValue := Items[i].FInitialValue + BooToChar[TRxCheckListBox(WinCtrl).Checked[j]];
   {$ENDIF}
   end;
   end;
end;

procedure TEditableCtrlHolder.RestoreInitialValues;
var
   i, j: byte;
   WinCtrl: TWinControl;
begin
   for i := 0 to Items.Count -1 do begin
   WinCtrl := Items[i].FWinControl;
   if (WinCtrl is TCustomEdit) then begin
      TCustomEdit(WinCtrl).Text := Items[i].FInitialValue;
   end else if (WinCtrl is TCustomComboBox) then begin
      TCustomComboBox(WinCtrl).ItemIndex := Items[i].FInitialValue;
   end else if (WinCtrl is TCheckBox) then begin
      TCheckBox(WinCtrl).Checked := Items[i].FInitialValue;
   {$IFDEF RxLib}
   end else if (WinCtrl is TRxCheckListBox) then begin
//      Items[i].FInitialValue := '';
      for j := 0 to TRxCheckListBox(WinCtrl).Items.Count -1 do
//         Items[i].FInitialValue := Items[i].FInitialValue + BooToSt[TRxCheckListBox(WinCtrl).Checked[j]];
         TRxCheckListBox(WinCtrl).Checked[j] := CharToBoo[VarToStr(Items[i].FInitialValue)[j+1]]; // := Items[i].FInitialValue + BooToSt[];
   {$ENDIF}
   end;
   end;
end;

procedure TEditableCtrlHolder.SetItems(Value: TEditableCtrlCollection);
begin
  FItems.Assign(Value);
end;

function TEditableCtrlHolder.Validate: TValidateResult;
function Visible(aWinControl: TWinControl): boolean;
begin
   result := aWinControl.Visible;
   if aWinControl.HasParent then
      result := result and Visible(aWinControl.Parent);
end;
var
   i: byte;
begin
   result := vrOk;
   FUnValidItem := nil;
   for i := 0 to Items.Count -1 do begin
//         OutputDebugString(PChar('i='+IntToStr(i)));
//      if not (Items[i].WinControl.Visible) or not (Items[i].WinControl.Parent.Visible) then
      if not Visible(Items[i].WinControl) then
         Continue;
      result := Items[i].ValidateEntry;
      if not (result = vrOk) then begin
         FUnValidItem := Items[i];
         break;
      end;
//      end;
   end;
end;

end.
